@@ -1,4 +1,3 @@
-use std::collections::HashSet;
 use std::collections::VecDeque;
 use std::io::Write;
 
@@ -18,6 +17,7 @@ struct State {
     pull: bool,
     push: bool,
     fixup: git_stack::config::Fixup,
+    repair: bool,
     dry_run: bool,
     snapshot_capacity: Option<usize>,
     protect_commit_count: Option<usize>,
@@ -60,6 +60,20 @@ impl State {
                 no_op
             }
         };
+        let repair = match (args.repair(), args.rebase) {
+            (Some(repair), _) => repair,
+            (_, true) => repo_config.auto_repair(),
+            _ => {
+                // Assume the user is only wanting to show the tree and not modify it.
+                if repo_config.auto_repair() {
+                    log::trace!(
+                        "Ignoring `auto-repair={}` without an explicit `--rebase`",
+                        repo_config.auto_repair()
+                    );
+                }
+                false
+            }
+        };
         let push = args.push;
         let protected = git_stack::git::ProtectedBranches::new(
             repo_config.protected_branches().iter().map(|s| s.as_str()),
@@ -91,6 +105,7 @@ impl State {
             .map(|name| resolve_explicit_base(&repo, name))
             .transpose()
             .with_code(proc_exit::Code::USAGE_ERR)?;
+
         let stacks = match (base, onto, repo_config.stack()) {
             (Some(base), None, git_stack::config::Stack::All) => {
                 let onto = base.clone();
@@ -180,6 +195,7 @@ impl State {
             pull,
             push,
             fixup,
+            repair,
             dry_run,
             snapshot_capacity,
             protect_commit_count,
@@ -204,6 +220,7 @@ impl State {
     }
 }
 
+#[derive(Debug)]
 struct StackState {
     base: git_stack::git::Branch,
     onto: git_stack::git::Branch,
@@ -244,17 +261,8 @@ pub fn stack(
     let repo = git2::Repository::discover(&cwd).with_code(proc_exit::Code::USAGE_ERR)?;
     let repo = git_stack::git::GitRepo::new(repo);
     let mut state = State::new(repo, args)?;
-    let mut stash_id = None;
 
     if state.pull {
-        if stash_id.is_none() {
-            stash_id = git_stack::git::stash_push(&mut state.repo, "branch-stash");
-        }
-        if state.repo.is_dirty() {
-            git_stack::git::stash_pop(&mut state.repo, stash_id);
-            return Err(proc_exit::Code::USAGE_ERR.with_message("Working tree is dirty, aborting"));
-        }
-
         // Update status of remote unprotected branches
         let mut push_branches: Vec<_> = state
             .stacks
@@ -266,7 +274,7 @@ pub fn stack(
             .collect();
         push_branches.sort_unstable();
         if !push_branches.is_empty() {
-            match git_fetch(&mut state.repo, &push_branches, state.dry_run) {
+            match git_prune_development(&mut state.repo, &push_branches, state.dry_run) {
                 Ok(_) => (),
                 Err(err) => {
                     log::warn!("Skipping fetch of `{}`, {}", state.repo.push_remote(), err);
@@ -274,20 +282,10 @@ pub fn stack(
             }
         }
 
-        let mut pulled_ids = HashSet::new();
         for stack in state.stacks.iter() {
-            let mut stack_pulled_ids = HashSet::new();
             if state.protected_branches.contains_oid(stack.onto.id) {
-                match git_pull(&mut state.repo, stack.onto.name.as_str(), state.dry_run) {
-                    Ok(pull_range) => {
-                        stack_pulled_ids.extend(
-                            state
-                                .repo
-                                .commits_from(pull_range.1)
-                                .take_while(|c| c.id != pull_range.0)
-                                .map(|c| c.id),
-                        );
-                    }
+                match git_fetch_upstream(&mut state.repo, stack.onto.name.as_str()) {
+                    Ok(_) => (),
                     Err(err) => {
                         log::warn!("Skipping pull of `{}`, {}", stack.onto.name, err);
                     }
@@ -298,38 +296,26 @@ pub fn stack(
                     stack.onto.name
                 );
             }
-            if !stack_pulled_ids.is_empty() {
-                match drop_branches(
-                    &mut state.repo,
-                    stack_pulled_ids.difference(&pulled_ids).cloned(),
-                    &stack.onto.name,
-                    &state.branches,
-                    &state.protected_branches,
-                    state.dry_run,
-                ) {
-                    Ok(()) => {}
-                    Err(err) => {
-                        log::warn!("Could not remove branches obsoleted by pull: {}", err);
-                    }
-                }
-                pulled_ids.extend(stack_pulled_ids);
-            }
-        }
-        if !pulled_ids.is_empty() {
-            state.update().with_code(proc_exit::Code::FAILURE)?;
         }
+        state.update().with_code(proc_exit::Code::FAILURE)?;
     }
 
     const STASH_STACK_NAME: &str = "git-stack";
     let mut success = true;
     let mut backed_up = false;
-    if state.rebase || state.fixup != git_stack::config::Fixup::Ignore {
-        if stash_id.is_none() {
+    let mut stash_id = None;
+    if state.rebase || state.fixup != git_stack::config::Fixup::Ignore || state.repair {
+        if stash_id.is_none() && !state.dry_run {
             stash_id = git_stack::git::stash_push(&mut state.repo, "branch-stash");
         }
         if state.repo.is_dirty() {
-            git_stack::git::stash_pop(&mut state.repo, stash_id);
-            return Err(proc_exit::Code::USAGE_ERR.with_message("Working tree is dirty, aborting"));
+            let message = "Working tree is dirty, aborting";
+            if state.dry_run {
+                log::error!("{}", message);
+            } else {
+                git_stack::git::stash_pop(&mut state.repo, stash_id);
+                return Err(proc_exit::Code::USAGE_ERR.with_message(message));
+            }
         }
 
         let mut snapshots = git_stack::stash::Stack::new(STASH_STACK_NAME, &state.repo);
@@ -389,7 +375,18 @@ pub fn stack(
     git_stack::git::stash_pop(&mut state.repo, stash_id);
 
     if backed_up {
-        log::info!("To undo, run `git branch-stash pop {}`", STASH_STACK_NAME);
+        let palette_stderr = if colored_stderr {
+            Palette::colored()
+        } else {
+            Palette::plain()
+        };
+        log::info!(
+            "{}",
+            palette_stderr.hint.paint(format_args!(
+                "To undo, run `git branch-stash pop {}`",
+                STASH_STACK_NAME
+            ))
+        );
     }
 
     if !success {
@@ -400,6 +397,7 @@ pub fn stack(
 }
 
 fn plan_changes(state: &State, stack: &StackState) -> eyre::Result<git_stack::git::Script> {
+    log::trace!("Planning stack changes with base={}", stack.base.name,);
     let graphed_branches = stack.graphed_branches();
     let base_commit = state
         .repo
@@ -407,7 +405,18 @@ fn plan_changes(state: &State, stack: &StackState) -> eyre::Result<git_stack::gi
         .expect("base branch is valid");
     let mut graph = git_stack::graph::Graph::from_branches(&state.repo, graphed_branches)?;
     graph.insert(&state.repo, git_stack::graph::Node::new(base_commit))?;
+    for branch in state.protected_branches.iter().flat_map(|(_, b)| b) {
+        if let Some(pull_id) = branch.pull_id {
+            let pull_commit = state
+                .repo
+                .find_commit(pull_id)
+                .expect("base branch is valid");
+            graph.insert(&state.repo, git_stack::graph::Node::new(pull_commit))?;
+        }
+    }
     git_stack::graph::protect_branches(&mut graph, &state.repo, &state.protected_branches);
+    let bases = git_stack::git::Branches::new([stack.base.clone(), stack.onto.clone()]);
+    git_stack::graph::protect_branches(&mut graph, &state.repo, &bases);
     if let Some(protect_commit_count) = state.protect_commit_count {
         git_stack::graph::protect_large_branches(&mut graph, protect_commit_count);
     }
@@ -416,13 +425,47 @@ fn plan_changes(state: &State, stack: &StackState) -> eyre::Result<git_stack::gi
         git_stack::graph::protect_foreign_branches(&mut graph, &user);
     }
 
+    let mut dropped_branches = Vec::new();
     if state.rebase {
-        git_stack::graph::rebase_branches(&mut graph, stack.onto.id);
-        git_stack::graph::drop_by_tree_id(&mut graph);
+        log::trace!("Rebasing onto {}", stack.onto.name);
+        let onto_id = stack.onto.pull_id.unwrap_or(stack.onto.id);
+        let pull_start_id = stack.onto.id;
+        let pull_start_id = state
+            .repo
+            .merge_base(pull_start_id, onto_id)
+            .unwrap_or(onto_id);
+
+        git_stack::graph::rebase_development_branches(&mut graph, onto_id);
+        git_stack::graph::rebase_pulled_branches(&mut graph, pull_start_id, onto_id);
+
+        let pull_range: Vec<_> = state
+            .repo
+            .commits_from(onto_id)
+            .take_while(|c| c.id != pull_start_id)
+            .collect();
+        git_stack::graph::drop_squashed_by_tree_id(
+            &mut graph,
+            pull_range.iter().map(|c| c.tree_id),
+        );
+        dropped_branches.extend(git_stack::graph::drop_merged_branches(
+            &mut graph,
+            pull_range.iter().map(|c| c.id),
+            &state.protected_branches,
+        ));
     }
     git_stack::graph::fixup(&mut graph, state.fixup);
+    if state.repair {
+        log::trace!("Repairing");
+        git_stack::graph::merge_stacks(&mut graph);
+        git_stack::graph::realign_stacks(&mut graph);
+    }
 
-    let script = git_stack::graph::to_script(&graph);
+    let mut script = git_stack::graph::to_script(&graph);
+    script.commands.extend(
+        dropped_branches
+            .into_iter()
+            .map(git_stack::git::Command::DeleteBranch),
+    );
 
     Ok(script)
 }
@@ -476,13 +519,27 @@ fn show(state: &State, colored_stdout: bool, colored_stderr: bool) -> eyre::Resu
             }
         }
 
+        log::trace!("Rendering stack base={}", stack.base.name,);
         let base_commit = state
             .repo
             .find_commit(stack.base.id)
             .expect("base branch is valid");
         let mut graph = git_stack::graph::Graph::from_branches(&state.repo, graphed_branches)?;
         graph.insert(&state.repo, git_stack::graph::Node::new(base_commit))?;
+        for branch in state.protected_branches.iter().flat_map(|(_, b)| b) {
+            if let Some(pull_id) = branch.pull_id {
+                if state.repo.merge_base(pull_id, graph.root_id()) == Some(graph.root_id()) {
+                    let pull_commit = state
+                        .repo
+                        .find_commit(pull_id)
+                        .expect("base branch is valid");
+                    graph.insert(&state.repo, git_stack::graph::Node::new(pull_commit))?;
+                }
+            }
+        }
         git_stack::graph::protect_branches(&mut graph, &state.repo, &state.protected_branches);
+        let bases = git_stack::git::Branches::new([stack.base.clone(), stack.onto.clone()]);
+        git_stack::graph::protect_branches(&mut graph, &state.repo, &bases);
         if let Some(protect_commit_count) = state.protect_commit_count {
             let protected =
                 git_stack::graph::protect_large_branches(&mut graph, protect_commit_count);
@@ -514,10 +571,38 @@ fn show(state: &State, colored_stdout: bool, colored_stderr: bool) -> eyre::Resu
         if state.dry_run {
             // Show as-if we performed all mutations
             if state.rebase {
-                git_stack::graph::rebase_branches(&mut graph, stack.onto.id);
-                git_stack::graph::drop_by_tree_id(&mut graph);
+                log::trace!("Rebasing onto {}", stack.onto.name);
+                let onto_id = stack.onto.pull_id.unwrap_or(stack.onto.id);
+                let pull_start_id = stack.onto.id;
+                let pull_start_id = state
+                    .repo
+                    .merge_base(pull_start_id, onto_id)
+                    .unwrap_or(onto_id);
+
+                git_stack::graph::rebase_development_branches(&mut graph, onto_id);
+                git_stack::graph::rebase_pulled_branches(&mut graph, pull_start_id, onto_id);
+
+                let pull_range: Vec<_> = state
+                    .repo
+                    .commits_from(onto_id)
+                    .take_while(|c| c.id != pull_start_id)
+                    .collect();
+                git_stack::graph::drop_squashed_by_tree_id(
+                    &mut graph,
+                    pull_range.iter().map(|c| c.tree_id),
+                );
+                git_stack::graph::drop_merged_branches(
+                    &mut graph,
+                    pull_range.iter().map(|c| c.id),
+                    &state.protected_branches,
+                );
             }
             git_stack::graph::fixup(&mut graph, state.fixup);
+            if state.repair {
+                log::trace!("Repairing");
+                git_stack::graph::merge_stacks(&mut graph);
+                git_stack::graph::realign_stacks(&mut graph);
+            }
         }
 
         git_stack::graph::pushable(&mut graph);
@@ -525,6 +610,7 @@ fn show(state: &State, colored_stdout: bool, colored_stderr: bool) -> eyre::Resu
         graphs.push(graph);
     }
     if graphs.is_empty() {
+        log::trace!("Rendering empty stack base={}", state.head_commit.id);
         let graph =
             git_stack::graph::Graph::new(git_stack::graph::Node::new(state.head_commit.clone()));
         graphs.push(graph);
@@ -610,17 +696,16 @@ fn resolve_implicit_base(
     Ok(branch.clone())
 }
 
-fn git_fetch(
+fn git_prune_development(
     repo: &mut git_stack::git::GitRepo,
     branches: &[&str],
     dry_run: bool,
 ) -> eyre::Result<()> {
-    let remote = repo.push_remote();
-    log::debug!("git fetch {} {}", remote, branches.join(" "));
-    if dry_run {
+    if branches.is_empty() {
         return Ok(());
     }
 
+    let remote = repo.push_remote();
     let output = std::process::Command::new("git")
         .arg("ls-remote")
         .arg("--heads")
@@ -634,6 +719,7 @@ fn git_fetch(
         eyre::bail!("Could not run `git fetch`");
     }
     let stdout = String::from_utf8(output.stdout).wrap_err("Could not run `git fetch`")?;
+    #[allow(clippy::needless_collect)]
     let remote_branches: Vec<_> = stdout
         .lines()
         .filter_map(|l| l.rsplit_once('/'))
@@ -644,254 +730,33 @@ fn git_fetch(
         if !remote_branches.contains(branch) {
             let remote_branch = format!("{}/{}", remote, branch);
             log::info!("Pruning {}", remote_branch);
-            let mut branch = repo
-                .raw()
-                .find_branch(&remote_branch, git2::BranchType::Remote)?;
-            branch.delete()?;
+            if !dry_run {
+                let mut branch = repo
+                    .raw()
+                    .find_branch(&remote_branch, git2::BranchType::Remote)?;
+                branch.delete()?;
+            }
         }
     }
 
-    // A little uncertain about some of the weirder authentication needs, just deferring to `git`
-    // instead of using `libgit2`
-    let status = std::process::Command::new("git")
-        .arg("fetch")
-        .arg(remote)
-        .args(remote_branches)
-        .status()
-        .wrap_err("Could not run `git fetch`")?;
-    if !status.success() {
-        eyre::bail!("`git fetch {} {}` failed", remote, branches.join(" "));
-    }
-
     Ok(())
 }
 
-fn git_pull(
-    repo: &mut git_stack::git::GitRepo,
-    branch_name: &str,
-    dry_run: bool,
-) -> eyre::Result<(git2::Oid, git2::Oid)> {
+fn git_fetch_upstream(repo: &mut git_stack::git::GitRepo, branch_name: &str) -> eyre::Result<()> {
     let remote = repo.pull_remote();
-    log::debug!("git pull --rebase {} {}", remote, branch_name);
-    let remote_branch_name = format!("{}/{}", remote, branch_name);
-    if dry_run {
-        let branch_id = repo.find_local_branch(branch_name).unwrap().id;
-        return Ok((branch_id, branch_id));
+    log::debug!("git fetch {} {}", remote, branch_name);
+    // A little uncertain about some of the weirder authentication needs, just deferring to `git`
+    // instead of using `libgit2`
+    let status = std::process::Command::new("git")
+        .arg("fetch")
+        .arg(remote)
+        .arg(branch_name)
+        .status()
+        .wrap_err("Could not run `git fetch`")?;
+    if !status.success() {
+        eyre::bail!("`git fetch {} {}` failed", remote, branch_name,);
     }
 
-    let pulled_range;
-    let mut tip_id;
-    {
-        // A little uncertain about some of the weirder authentication needs, just deferring to `git`
-        // instead of using `libgit2`
-        let status = std::process::Command::new("git")
-            .arg("fetch")
-            .arg(remote)
-            .arg(branch_name)
-            .status()
-            .wrap_err("Could not run `git fetch`")?;
-        if !status.success() {
-            eyre::bail!("`git fetch {} {}` failed", remote, branch_name,);
-        }
-
-        let local_branch = repo
-            .raw()
-            .find_branch(branch_name, git2::BranchType::Local)
-            .wrap_err_with(|| eyre::eyre!("local branch `{}` doesn't exist", branch_name))?;
-        let local_branch_annotated = {
-            repo.raw()
-                .reference_to_annotated_commit(local_branch.get())?
-        };
-        log::trace!(
-            "rebase local {}={}",
-            branch_name,
-            local_branch_annotated.id()
-        );
-
-        let remote_branch = repo
-            .raw()
-            .find_branch(&remote_branch_name, git2::BranchType::Remote)
-            .wrap_err_with(|| {
-                eyre::eyre!("remote branch `{}` doesn't exist", remote_branch_name)
-            })?;
-        let remote_branch_annotated = repo
-            .raw()
-            .reference_to_annotated_commit(remote_branch.get())?;
-        log::trace!(
-            "rebase remote {}={}",
-            remote_branch_name,
-            remote_branch_annotated.id()
-        );
-        let end_id = remote_branch_annotated.id();
-        tip_id = end_id;
-
-        let base_id = repo
-            .merge_base(local_branch_annotated.id(), remote_branch_annotated.id())
-            .ok_or_else(|| {
-                eyre::eyre!(
-                    "No common base between {} and {}",
-                    branch_name,
-                    remote_branch_name
-                )
-            })?;
-        let base_annotated = repo.raw().find_annotated_commit(base_id).unwrap();
-        log::trace!("rebase base {}", base_annotated.id());
-
-        let merge_base_id =
-            repo.merge_base(local_branch_annotated.id(), remote_branch_annotated.id());
-        if merge_base_id == Some(remote_branch_annotated.id()) {
-            log::debug!("{} is up-to-date with {}", branch_name, remote_branch_name);
-            return Ok((local_branch_annotated.id(), local_branch_annotated.id()));
-        }
-        let start_id = merge_base_id.unwrap_or(end_id);
-        pulled_range = (start_id, end_id);
-
-        let mut rebase = repo
-            .raw()
-            .rebase(
-                Some(&local_branch_annotated),
-                Some(&base_annotated),
-                Some(&remote_branch_annotated),
-                Some(git2::RebaseOptions::new().inmemory(true)),
-            )
-            .wrap_err_with(|| {
-                eyre::eyre!(
-                    "failed to rebase `{}` onto `{}`",
-                    branch_name,
-                    remote_branch_name
-                )
-            })?;
-
-        while let Some(op) = rebase.next() {
-            let op = op
-                .map_err(|e| {
-                    let _ = rebase.abort();
-                    e
-                })
-                .wrap_err_with(|| {
-                    eyre::eyre!(
-                        "failed to rebase `{}` onto `{}`",
-                        branch_name,
-                        remote_branch_name
-                    )
-                })?;
-            log::trace!("Rebase: {:?} {}", op.kind(), op.id());
-            if rebase.inmemory_index().unwrap().has_conflicts() {
-                eyre::bail!(
-                    "conflicts between {} and {}",
-                    branch_name,
-                    remote_branch_name
-                );
-            }
-
-            let sig = repo.raw().signature().unwrap();
-            let commit_id = rebase
-                .commit(None, &sig, None)
-                .map_err(|e| {
-                    let _ = rebase.abort();
-                    e
-                })
-                .wrap_err_with(|| {
-                    eyre::eyre!(
-                        "failed to rebase `{}` onto `{}`",
-                        branch_name,
-                        remote_branch_name
-                    )
-                })?;
-            tip_id = commit_id;
-        }
-
-        rebase.finish(None).wrap_err_with(|| {
-            eyre::eyre!(
-                "failed to rebase `{}` onto `{}`",
-                branch_name,
-                remote_branch_name
-            )
-        })?;
-    }
-
-    let head_branch = repo.head_branch();
-    let head_branch_name = head_branch.as_ref().map(|b| b.name.as_str());
-    if head_branch_name == Some(branch_name) {
-        log::trace!("Updating {} (HEAD)", branch_name);
-        repo.detach().wrap_err_with(|| {
-            eyre::eyre!(
-                "failed to update `{}` to `{}`",
-                branch_name,
-                remote_branch_name
-            )
-        })?;
-        repo.branch(branch_name, tip_id).wrap_err_with(|| {
-            eyre::eyre!(
-                "failed to update `{}` to `{}`",
-                branch_name,
-                remote_branch_name
-            )
-        })?;
-        repo.switch(branch_name).wrap_err_with(|| {
-            eyre::eyre!(
-                "failed to update `{}` to `{}`",
-                branch_name,
-                remote_branch_name
-            )
-        })?;
-    } else {
-        log::trace!("Updating {}", branch_name);
-        repo.branch(branch_name, tip_id).wrap_err_with(|| {
-            eyre::eyre!(
-                "failed to update `{}` to `{}`",
-                branch_name,
-                remote_branch_name
-            )
-        })?;
-    }
-
-    Ok(pulled_range)
-}
-
-fn drop_branches(
-    repo: &mut git_stack::git::GitRepo,
-    commit_ids: impl Iterator<Item = git2::Oid>,
-    potential_head: &str,
-    branches: &git_stack::git::Branches,
-    protected_branches: &git_stack::git::Branches,
-    dry_run: bool,
-) -> eyre::Result<()> {
-    let head_branch = repo.head_branch();
-    let head_branch_name = head_branch.as_ref().map(|b| b.name.as_str());
-
-    for commit_id in commit_ids {
-        let commit_branches: HashSet<_> = branches.get(commit_id).into_iter().flatten().collect();
-        let commit_protected_branches: HashSet<_> = protected_branches
-            .get(commit_id)
-            .into_iter()
-            .flatten()
-            .collect();
-        let mut commit_unprotected: Vec<_> = commit_branches
-            .difference(&commit_protected_branches)
-            .collect();
-        commit_unprotected.sort_unstable();
-        for branch in commit_unprotected {
-            if branch.name == potential_head {
-                continue;
-            } else if head_branch_name == Some(branch.name.as_str()) {
-                // Don't leave HEAD detached but instead switch to the branch we pulled
-                log::trace!("git switch {}", potential_head);
-                if !dry_run {
-                    repo.switch(potential_head)?;
-                }
-                log::trace!("git branch -D {}", branch.name);
-                if !dry_run {
-                    repo.delete_branch(&branch.name)?;
-                }
-            } else {
-                log::trace!("git branch -D {}", branch.name);
-                if !dry_run {
-                    repo.delete_branch(&branch.name)?;
-                }
-            }
-        }
-    }
     Ok(())
 }
 
@@ -1017,30 +882,30 @@ impl<'r> std::fmt::Display for DisplayTree<'r> {
 
         let is_visible: Box<dyn Fn(&git_stack::graph::Node) -> bool> = match self.show {
             git_stack::config::Format::Silent => unreachable!("No silent view for tree"),
-            git_stack::config::Format::Commits => Box::new(|_| false),
+            git_stack::config::Format::Commits => Box::new(|_| true),
             git_stack::config::Format::BranchCommits => Box::new(|node| {
-                let interesting_commit =
-                    node.commit.id == head_branch.id || node.commit.id == self.graph.root_id();
+                let interesting_commit = node.commit.id == head_branch.id
+                    || node.commit.id == self.graph.root_id()
+                    || node.children.is_empty();
                 let boring_commit = node.branches.is_empty() && node.children.len() == 1;
                 let protected = node.action.is_protected();
                 interesting_commit || !boring_commit || !protected
             }),
             git_stack::config::Format::Branches => Box::new(|node| {
-                let interesting_commit =
-                    node.commit.id == head_branch.id || node.commit.id == self.graph.root_id();
+                let interesting_commit = node.commit.id == head_branch.id
+                    || node.commit.id == self.graph.root_id()
+                    || node.children.is_empty();
                 let boring_commit = node.branches.is_empty() && node.children.len() == 1;
                 interesting_commit || !boring_commit
             }),
             git_stack::config::Format::Debug => unreachable!("No debug view for tree"),
         };
 
-        let mut tree = to_tree(
+        let mut tree = node_to_tree(
             self.repo,
             &head_branch,
-            &self.protected_branches,
             self.graph,
             self.graph.root_id(),
-            &self.palette,
             &is_visible,
         );
         if self.stacked {
@@ -1048,21 +913,24 @@ impl<'r> std::fmt::Display for DisplayTree<'r> {
         } else {
             tree.sort();
         }
-        let tree = tree.into_display();
+        let tree = tree.into_display(
+            self.repo,
+            &head_branch,
+            &self.protected_branches,
+            &self.palette,
+        );
         tree.fmt(f)
     }
 }
 
-fn to_tree<'r>(
+fn node_to_tree<'r>(
     repo: &'r git_stack::git::GitRepo,
     head_branch: &'r git_stack::git::Branch,
-    protected_branches: &'r git_stack::git::Branches,
     graph: &'r git_stack::graph::Graph,
     mut node_id: git2::Oid,
-    palette: &'r Palette,
     is_visible: &dyn Fn(&git_stack::graph::Node) -> bool,
 ) -> Tree<'r> {
-    loop {
+    for ellide_count in 0.. {
         let node = graph.get(node_id).expect("all children exist");
         // The API requires us to handle 0 or many children, so not checking visibility
         if node.children.len() == 1 && !is_visible(node) {
@@ -1070,46 +938,102 @@ fn to_tree<'r>(
             continue;
         }
 
-        let mut weight = if node.action.is_protected() {
-            Weight::Protected(0)
-        } else if node.commit.id == head_branch.id {
-            Weight::Head(0)
-        } else {
-            Weight::Commit(0)
+        let mut tree = Tree {
+            root: node,
+            weight: default_weight(node, head_branch),
+            stacks: Default::default(),
         };
 
-        let mut stacks = Vec::new();
-        for child_id in node.children.iter().copied() {
-            let child_tree = to_tree(
-                repo,
-                head_branch,
-                protected_branches,
-                graph,
-                child_id,
-                palette,
-                is_visible,
-            );
-            weight = weight.max(child_tree.weight);
-            stacks.push(vec![child_tree]);
-        }
+        append_children(&mut tree, repo, head_branch, graph, node, is_visible);
+
+        tree.weight += ellide_count;
 
-        let tree = Tree {
-            root: RenderNode {
-                repo,
-                head_branch,
-                protected_branches,
-                node: Some(node),
-                palette,
-            },
-            weight,
-            stacks,
-        };
         return tree;
     }
+
+    unreachable!("above loop always hits `return`")
+}
+
+fn append_children<'r>(
+    tree: &mut Tree<'r>,
+    repo: &'r git_stack::git::GitRepo,
+    head_branch: &'r git_stack::git::Branch,
+    graph: &'r git_stack::graph::Graph,
+    mut parent_node: &'r git_stack::graph::Node,
+    is_visible: &dyn Fn(&git_stack::graph::Node) -> bool,
+) {
+    match parent_node.children.len() {
+        0 => {}
+        1 => {
+            for linear_count in 1.. {
+                let node_id = *parent_node.children.iter().next().unwrap();
+                let node = graph.get(node_id).expect("all children exist");
+                match node.children.len() {
+                    0 => {
+                        let child_tree = Tree {
+                            root: node,
+                            weight: default_weight(node, head_branch),
+                            stacks: Default::default(),
+                        };
+                        tree.weight = tree.weight.max(child_tree.weight + linear_count);
+                        if tree.stacks.is_empty() {
+                            tree.stacks.push(Vec::new());
+                        }
+                        tree.stacks[0].push(child_tree);
+                        break;
+                    }
+                    1 => {
+                        if is_visible(node) {
+                            let child_tree = Tree {
+                                root: node,
+                                weight: default_weight(node, head_branch),
+                                stacks: Default::default(),
+                            };
+                            // `tree.weight`: rely on a terminating case for updating
+                            if tree.stacks.is_empty() {
+                                tree.stacks.push(Vec::new());
+                            }
+                            tree.stacks[0].push(child_tree);
+                        }
+                        parent_node = node;
+                        continue;
+                    }
+                    _ => {
+                        let child_tree =
+                            node_to_tree(repo, head_branch, graph, node_id, is_visible);
+                        tree.weight = tree.weight.max(child_tree.weight + linear_count);
+                        if tree.stacks.is_empty() {
+                            tree.stacks.push(Vec::new());
+                        }
+                        tree.stacks[0].push(child_tree);
+                        break;
+                    }
+                }
+            }
+        }
+        _ => {
+            for child_id in parent_node.children.iter().copied() {
+                let child_tree = node_to_tree(repo, head_branch, graph, child_id, is_visible);
+                tree.weight = tree.weight.max(child_tree.weight + 1);
+                tree.stacks.push(vec![child_tree]);
+            }
+        }
+    }
 }
 
+fn default_weight(node: &git_stack::graph::Node, head_branch: &git_stack::git::Branch) -> Weight {
+    if node.action.is_protected() {
+        Weight::Protected(0)
+    } else if node.commit.id == head_branch.id {
+        Weight::Head(0)
+    } else {
+        Weight::Commit(0)
+    }
+}
+
+#[derive(Debug)]
 struct Tree<'r> {
-    root: RenderNode<'r>,
+    root: &'r git_stack::graph::Node,
     stacks: Vec<Vec<Self>>,
     weight: Weight,
 }
@@ -1142,21 +1066,64 @@ impl<'r> Tree<'r> {
         }
     }
 
-    fn into_display(self) -> termtree::Tree<RenderNode<'r>> {
-        let mut tree = termtree::Tree::root(self.root);
-        if self.stacks.len() == 1 {
-            for stack in self.stacks.into_iter() {
-                for child in stack.into_iter() {
-                    tree.push(child.into_display());
-                }
-            }
-        } else {
-            for stack in self.stacks.into_iter() {
-                let mut stack_tree = termtree::Tree::root(self.root.joint());
-                for child in stack.into_iter() {
-                    stack_tree.push(child.into_display());
+    fn into_display(
+        self,
+        repo: &'r git_stack::git::GitRepo,
+        head_branch: &'r git_stack::git::Branch,
+        protected_branches: &'r git_stack::git::Branches,
+        palette: &'r Palette,
+    ) -> termtree::Tree<RenderNode<'r>> {
+        let root = RenderNode {
+            repo,
+            head_branch,
+            protected_branches,
+            node: Some(self.root),
+            palette,
+        };
+        let mut tree = termtree::Tree::root(root).with_glyphs(GLYPHS);
+        let joint = RenderNode {
+            repo,
+            head_branch,
+            protected_branches,
+            node: None,
+            palette,
+        };
+        let stacks_len = self.stacks.len();
+        for (i, stack) in self.stacks.into_iter().enumerate() {
+            if i < stacks_len - 1 {
+                let mut stack_tree = termtree::Tree::root(joint).with_glyphs(JOINT_GLYPHS);
+                for child_tree in stack.into_iter() {
+                    stack_tree.push(child_tree.into_display(
+                        repo,
+                        head_branch,
+                        protected_branches,
+                        palette,
+                    ));
                 }
                 tree.push(stack_tree);
+            } else {
+                for child_tree in stack.into_iter() {
+                    let child = RenderNode {
+                        repo,
+                        head_branch,
+                        protected_branches,
+                        node: Some(child_tree.root),
+                        palette,
+                    };
+                    tree.push(termtree::Tree::root(child).with_glyphs(GLYPHS));
+                    for child_stack in child_tree.stacks.into_iter() {
+                        let mut stack_tree = termtree::Tree::root(joint).with_glyphs(JOINT_GLYPHS);
+                        for child_tree in child_stack.into_iter() {
+                            stack_tree.push(child_tree.into_display(
+                                repo,
+                                head_branch,
+                                protected_branches,
+                                palette,
+                            ));
+                        }
+                        tree.push(stack_tree);
+                    }
+                }
             }
         }
         tree
@@ -1184,7 +1151,25 @@ impl Weight {
     }
 }
 
-#[derive(Copy, Clone)]
+impl std::ops::Add<usize> for Weight {
+    type Output = Self;
+
+    fn add(self, other: usize) -> Self {
+        match self {
+            Self::Protected(s) => Self::Protected(s.saturating_add(other)),
+            Self::Head(s) => Self::Head(s.saturating_add(other)),
+            Self::Commit(s) => Self::Commit(s.saturating_add(other)),
+        }
+    }
+}
+
+impl std::ops::AddAssign<usize> for Weight {
+    fn add_assign(&mut self, other: usize) {
+        *self = *self + other;
+    }
+}
+
+#[derive(Copy, Clone, Debug)]
 struct RenderNode<'r> {
     repo: &'r git_stack::git::GitRepo,
     head_branch: &'r git_stack::git::Branch,
@@ -1193,17 +1178,18 @@ struct RenderNode<'r> {
     palette: &'r Palette,
 }
 
-impl<'r> RenderNode<'r> {
-    fn joint(&self) -> Self {
-        Self {
-            repo: self.repo,
-            head_branch: self.head_branch,
-            protected_branches: self.protected_branches,
-            node: None,
-            palette: self.palette,
-        }
-    }
-}
+const GLYPHS: termtree::GlyphPalette = termtree::GlyphPalette {
+    middle_item: "⌽",
+    last_item: "⌽",
+    item_indent: " ",
+    skip_indent: " ",
+    ..termtree::GlyphPalette::new()
+};
+
+const JOINT_GLYPHS: termtree::GlyphPalette = termtree::GlyphPalette {
+    item_indent: "─┐",
+    ..termtree::GlyphPalette::new()
+};
 
 // Shared implementation doesn't mean shared requirements, we want to track according to
 // requirements
@@ -1277,8 +1263,6 @@ impl<'r> std::fmt::Display for RenderNode<'r> {
             } else {
                 write!(f, "{}", summary)?;
             }
-        } else {
-            write!(f, "o")?;
         }
         Ok(())
     }
