@@ -246,6 +246,7 @@ impl GitRepo {
     ) -> impl Iterator<Item = std::rc::Rc<Commit>> + '_ {
         let mut revwalk = self.repo.revwalk().unwrap();
         revwalk.push(head_id).unwrap();
+        revwalk.set_sorting(git2::Sort::TOPOLOGICAL).unwrap();
 
         revwalk
             .filter_map(Result::ok)
@@ -322,13 +323,14 @@ impl GitRepo {
         if base_id == head_id {
             return Ok(cherry_id);
         }
-        let base_commit = self.repo.find_annotated_commit(base_id)?;
-        let head_commit = self.repo.find_annotated_commit(head_id)?;
-        let cherry_commit = self.repo.find_annotated_commit(cherry_id)?;
+        let base_ann_commit = self.repo.find_annotated_commit(base_id)?;
+        let head_ann_commit = self.repo.find_annotated_commit(head_id)?;
+        let cherry_ann_commit = self.repo.find_annotated_commit(cherry_id)?;
+        let cherry_commit = self.repo.find_commit(cherry_id)?;
         let mut rebase = self.repo.rebase(
-            Some(&cherry_commit),
-            Some(&base_commit),
-            Some(&head_commit),
+            Some(&cherry_ann_commit),
+            Some(&base_ann_commit),
+            Some(&head_ann_commit),
             Some(git2::RebaseOptions::new().inmemory(true)),
         )?;
 
@@ -361,7 +363,11 @@ impl GitRepo {
                 ));
             }
 
-            let sig = self.repo.signature().unwrap();
+            let mut sig = self.repo.signature()?;
+            if let (Some(name), Some(email)) = (sig.name(), sig.email()) {
+                // For simple rebases, preserve the original commit time
+                sig = git2::Signature::new(name, email, &cherry_commit.time())?.to_owned();
+            }
             let commit_id = match rebase.commit(None, &sig, None).map_err(|e| {
                 let _ = rebase.abort();
                 e
@@ -595,6 +601,16 @@ impl GitRepo {
     }
 }
 
+impl std::fmt::Debug for GitRepo {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> Result<(), std::fmt::Error> {
+        f.debug_struct("GitRepo")
+            .field("repo", &self.repo.workdir())
+            .field("push_remote", &self.push_remote.as_deref())
+            .field("pull_remote", &self.pull_remote.as_deref())
+            .finish()
+    }
+}
+
 impl Repo for GitRepo {
     fn user(&self) -> Option<std::rc::Rc<str>> {
         self.user()
