@@ -1,3 +1,4 @@
+use std::collections::BTreeMap;
 use std::collections::HashSet;
 use std::collections::VecDeque;
 
@@ -93,7 +94,7 @@ fn protect_large_branches_recursive(
                 protect_large_branches_recursive(graph, child_id, count + 1, max, large_branches);
         }
         if needs_protection {
-            let mut node = graph.get_mut(node_id).expect("all children exist");
+            let node = graph.get_mut(node_id).expect("all children exist");
             node.action = crate::graph::Action::Protected;
         }
     } else {
@@ -108,7 +109,7 @@ fn mark_branch_protected(graph: &mut Graph, node_id: git2::Oid, branches: &mut V
     let mut protected_queue = VecDeque::new();
     protected_queue.push_back(node_id);
     while let Some(current_id) = protected_queue.pop_front() {
-        let mut current = graph.get_mut(current_id).expect("all children exist");
+        let current = graph.get_mut(current_id).expect("all children exist");
         current.action = crate::graph::Action::Protected;
 
         if current.branches.is_empty() {
@@ -306,7 +307,7 @@ fn is_personal_branch(graph: &Graph, node_id: git2::Oid, user: &str, ignore: &[g
 /// # Panics
 ///
 /// - If `new_base_id` doesn't exist
-pub fn rebase_branches(graph: &mut Graph, new_base_id: git2::Oid) {
+pub fn rebase_development_branches(graph: &mut Graph, new_base_id: git2::Oid) {
     debug_assert!(graph.get(new_base_id).is_some());
 
     let mut protected_queue = VecDeque::new();
@@ -344,6 +345,32 @@ pub fn rebase_branches(graph: &mut Graph, new_base_id: git2::Oid) {
     }
 }
 
+/// Update branches from `pull_start` to `pull_end`
+///
+/// A normal `rebase_development_branches` only looks at development commits.  If `main` is pristine or if the
+/// user has branches on the same commit as `main`, we should also update these to what we pulled.
+pub fn rebase_pulled_branches(graph: &mut Graph, pull_start: git2::Oid, pull_end: git2::Oid) {
+    if pull_start == pull_end {
+        return;
+    }
+
+    let mut branches = Default::default();
+    std::mem::swap(
+        &mut branches,
+        &mut graph
+            .get_mut(pull_start)
+            .expect("all children exist")
+            .branches,
+    );
+    std::mem::swap(
+        &mut branches,
+        &mut graph
+            .get_mut(pull_end)
+            .expect("all children exist")
+            .branches,
+    );
+}
+
 pub fn pushable(graph: &mut Graph) {
     let mut node_queue: VecDeque<(git2::Oid, Option<&str>)> = VecDeque::new();
 
@@ -392,55 +419,30 @@ pub fn pushable(graph: &mut Graph) {
 ///
 /// This assumes that the Node was rebased onto all of the new potentially squash-merged Nodes and
 /// we extract the potential tree_id's from those protected commits.
-pub fn drop_by_tree_id(graph: &mut Graph) {
-    let mut node_queue: VecDeque<(git2::Oid, HashSet<git2::Oid>)> = VecDeque::new();
-    if graph.root().action.is_protected() {
-        node_queue.push_back((graph.root_id(), HashSet::new()));
-    }
-    while let Some((protected_id, mut protected_tree_ids)) = node_queue.pop_front() {
-        protected_tree_ids.insert(
-            graph
-                .get(protected_id)
-                .expect("all children exist")
-                .commit
-                .tree_id,
-        );
+pub fn drop_squashed_by_tree_id(
+    graph: &mut Graph,
+    pulled_tree_ids: impl Iterator<Item = git2::Oid>,
+) {
+    let pulled_tree_ids: HashSet<_> = pulled_tree_ids.collect();
 
-        let protected_children = graph
-            .get(protected_id)
+    let mut protected_queue = VecDeque::new();
+    let root_action = graph.root().action;
+    if root_action.is_protected() {
+        protected_queue.push_back(graph.root_id());
+    }
+    while let Some(current_id) = protected_queue.pop_front() {
+        let current_children = graph
+            .get(current_id)
             .expect("all children exist")
             .children
             .clone();
-        match protected_children.len() {
-            0 => (),
-            1 => {
-                let child_id = protected_children.into_iter().next().unwrap();
-                let child_action = graph.get(child_id).expect("all children exist").action;
-                if child_action.is_protected() {
-                    node_queue.push_back((child_id, protected_tree_ids));
-                } else {
-                    drop_first_branch_by_tree_id(
-                        graph,
-                        child_id,
-                        HashSet::new(),
-                        protected_tree_ids,
-                    );
-                }
-            }
-            _ => {
-                for child_id in protected_children {
-                    let child_action = graph.get(child_id).expect("all children exist").action;
-                    if child_action.is_protected() {
-                        node_queue.push_back((child_id, protected_tree_ids.clone()));
-                    } else {
-                        drop_first_branch_by_tree_id(
-                            graph,
-                            child_id,
-                            HashSet::new(),
-                            protected_tree_ids.clone(),
-                        );
-                    }
-                }
+
+        for child_id in current_children {
+            let child_action = graph.get(child_id).expect("all children exist").action;
+            if child_action.is_protected() || child_action.is_delete() {
+                protected_queue.push_back(child_id);
+            } else {
+                drop_first_branch_by_tree_id(graph, child_id, HashSet::new(), &pulled_tree_ids);
             }
         }
     }
@@ -449,10 +451,9 @@ pub fn drop_by_tree_id(graph: &mut Graph) {
 fn drop_first_branch_by_tree_id(
     graph: &mut Graph,
     node_id: git2::Oid,
-    mut branch_ids: std::collections::HashSet<git2::Oid>,
-    protected_tree_ids: std::collections::HashSet<git2::Oid>,
+    mut branch_ids: HashSet<git2::Oid>,
+    pulled_tree_ids: &HashSet<git2::Oid>,
 ) {
-    #![allow(clippy::if_same_then_else)]
     branch_ids.insert(node_id);
 
     let node = graph.get(node_id).expect("all children exist");
@@ -467,12 +468,11 @@ fn drop_first_branch_by_tree_id(
     let node_tree_id = node.commit.tree_id;
 
     if is_branch {
-        if protected_tree_ids.contains(&node_tree_id) {
+        if pulled_tree_ids.contains(&node_tree_id) {
             for branch_id in branch_ids {
                 graph.get_mut(branch_id).expect("all children exist").action =
                     crate::graph::Action::Delete;
             }
-        } else {
         }
     } else {
         let node_children = graph
@@ -484,7 +484,7 @@ fn drop_first_branch_by_tree_id(
             0 => {}
             1 => {
                 let child_id = node_children.into_iter().next().unwrap();
-                drop_first_branch_by_tree_id(graph, child_id, branch_ids, protected_tree_ids);
+                drop_first_branch_by_tree_id(graph, child_id, branch_ids, pulled_tree_ids);
             }
             _ => {
                 for child_id in node_children {
@@ -492,7 +492,7 @@ fn drop_first_branch_by_tree_id(
                         graph,
                         child_id,
                         branch_ids.clone(),
-                        protected_tree_ids.clone(),
+                        pulled_tree_ids,
                     );
                 }
             }
@@ -500,6 +500,39 @@ fn drop_first_branch_by_tree_id(
     }
 }
 
+/// Drop branches merged among the pulled IDs
+///
+/// The removal in `graph` is purely superficial since nothing can act on it.  The returned branch
+/// names is the important part.
+pub fn drop_merged_branches(
+    graph: &mut Graph,
+    pulled_ids: impl Iterator<Item = git2::Oid>,
+    protected_branches: &crate::git::Branches,
+) -> Vec<String> {
+    let mut removed = Vec::new();
+    let protected_branch_names: HashSet<_> = protected_branches
+        .iter()
+        .flat_map(|(_, b)| b.iter())
+        .map(|b| b.name.as_str())
+        .collect();
+
+    for pulled_id in pulled_ids {
+        // HACK: Depending on how merges in master worked out, not all commits will be present
+        if let Some(node) = graph.get_mut(pulled_id) {
+            if !node.branches.is_empty() {
+                for i in (node.branches.len() - 1)..=0 {
+                    if !protected_branch_names.contains(node.branches[i].name.as_str()) {
+                        let branch = node.branches.remove(i);
+                        removed.push(branch.name);
+                    }
+                }
+            }
+        }
+    }
+
+    removed
+}
+
 pub fn fixup(graph: &mut Graph, effect: crate::config::Fixup) {
     if effect == crate::config::Fixup::Ignore {
         return;
@@ -512,7 +545,7 @@ pub fn fixup(graph: &mut Graph, effect: crate::config::Fixup) {
     }
     while let Some(current_id) = protected_queue.pop_front() {
         let current_children = graph
-            .get_mut(current_id)
+            .get(current_id)
             .expect("all children exist")
             .children
             .clone();
@@ -538,7 +571,7 @@ fn fixup_branch(
 
     let mut outstanding = std::collections::BTreeMap::new();
     let node_children = graph
-        .get_mut(node_id)
+        .get(node_id)
         .expect("all children exist")
         .children
         .clone();
@@ -552,7 +585,7 @@ fn fixup_branch(
                 for fixup_id in fixup_ids.iter().copied() {
                     let fixup = graph.get_mut(fixup_id).expect("all children exist");
                     assert!(fixup.action == crate::graph::Action::Pick);
-                    fixup.action = crate::graph::Action::Squash;
+                    fixup.action = crate::graph::Action::Fixup;
                 }
             }
             splice_after(graph, node_id, fixup_ids);
@@ -579,7 +612,7 @@ fn fixup_node(
     debug_assert_ne!(effect, crate::config::Fixup::Ignore);
 
     let node_children = graph
-        .get_mut(node_id)
+        .get(node_id)
         .expect("all children exist")
         .children
         .clone();
@@ -623,7 +656,7 @@ fn fixup_node(
             for fixup_id in fixup_ids.iter().copied() {
                 let fixup = graph.get_mut(fixup_id).expect("all children exist");
                 assert!(fixup.action == crate::graph::Action::Pick);
-                fixup.action = crate::graph::Action::Squash;
+                fixup.action = crate::graph::Action::Fixup;
             }
         }
         splice_after(graph, node_id, fixup_ids);
@@ -680,6 +713,173 @@ fn splice_after(graph: &mut Graph, node_id: git2::Oid, fixup_ids: Vec<git2::Oid>
     }
 }
 
+/// When a branch has extra commits, update dependent branches to the latest
+pub fn realign_stacks(graph: &mut Graph) {
+    let mut protected_queue = VecDeque::new();
+    let root_action = graph.root().action;
+    if root_action.is_protected() {
+        protected_queue.push_back(graph.root_id());
+    }
+    while let Some(current_id) = protected_queue.pop_front() {
+        let current_children = graph
+            .get(current_id)
+            .expect("all children exist")
+            .children
+            .clone();
+
+        for child_id in current_children {
+            let child_action = graph.get(child_id).expect("all children exist").action;
+            if child_action.is_protected() || child_action.is_delete() {
+                protected_queue.push_back(child_id);
+            } else {
+                realign_stack(graph, child_id);
+            }
+        }
+    }
+}
+
+fn realign_stack(graph: &mut Graph, node_id: git2::Oid) {
+    let mut children = std::collections::BTreeSet::new();
+
+    let mut current_id = node_id;
+    loop {
+        let current = graph.get_mut(current_id).expect("all children exist");
+        if current.branches.is_empty() {
+            let mut current_children: Vec<_> = current.children.iter().copied().collect();
+            match current_children.len() {
+                0 => {
+                    current.children.extend(children);
+                    return;
+                }
+                1 => {
+                    current_id = current_children.into_iter().next().unwrap();
+                }
+                _ => {
+                    // Assuming the more recent work is a continuation of the existing stack and
+                    // aligning the other stacks to be on top of it
+                    //
+                    // This should be safe in light of our rebases since we don't preserve the time
+                    current_children.sort_unstable_by_key(|id| {
+                        graph.get(*id).expect("all children exist").commit.time
+                    });
+                    let newest = current_children.pop().unwrap();
+                    {
+                        let current = graph.get_mut(current_id).expect("all children exist");
+                        for child_id in &current_children {
+                            current.children.remove(child_id);
+                        }
+                    }
+                    children.extend(current_children);
+                    current_id = newest;
+                }
+            }
+        } else {
+            if !children.is_empty() {
+                current.children.extend(children);
+            }
+            return;
+        }
+    }
+}
+
+/// When a rebase has split stack, re-combine them
+pub fn merge_stacks(graph: &mut Graph) {
+    let mut protected_queue = VecDeque::new();
+    let root_action = graph.root().action;
+    if root_action.is_protected() {
+        protected_queue.push_back(graph.root_id());
+    }
+    while let Some(current_id) = protected_queue.pop_front() {
+        let current_children = graph
+            .get(current_id)
+            .expect("all children exist")
+            .children
+            .clone();
+
+        let mut unprotected_children = CommitTimesByTreeId::new();
+        for child_id in current_children {
+            let child_action = graph.get(child_id).expect("all children exist").action;
+            if child_action.is_protected() || child_action.is_delete() {
+                protected_queue.push_back(child_id);
+            } else {
+                let child = graph.get(child_id).expect("all children exist");
+
+                unprotected_children
+                    .entry(child.commit.tree_id)
+                    .or_insert_with(Vec::new)
+                    .push((child.commit.time, child_id));
+            }
+        }
+        if !unprotected_children.is_empty() {
+            merge_child_stacks(graph, current_id, unprotected_children);
+        }
+    }
+}
+
+type CommitTimesByTreeId = BTreeMap<git2::Oid, Vec<(std::time::SystemTime, git2::Oid)>>;
+
+fn merge_child_stacks(graph: &mut Graph, node_id: git2::Oid, children: CommitTimesByTreeId) {
+    let mut queue = VecDeque::new();
+    queue.push_back((node_id, children));
+    while let Some((current_id, children)) = queue.pop_front() {
+        for (_, mut children) in children {
+            match children.len() {
+                0 => unreachable!("each entry has at least one commit"),
+                1 => {
+                    enqueue_merge_stack(&mut queue, graph, children[0].1);
+                }
+                _ => {
+                    children.sort_unstable();
+                    let last_index = children.len() - 1;
+                    let last_child_id = children[last_index].1;
+
+                    for (_, child_id) in &children[0..last_index] {
+                        let mut grandchildren = Default::default();
+                        let mut branches = Default::default();
+
+                        let child = graph.get_mut(*child_id).expect("all children exist");
+                        std::mem::swap(&mut child.branches, &mut branches);
+                        std::mem::swap(&mut child.children, &mut grandchildren);
+
+                        let last_child = graph.get_mut(last_child_id).expect("all children exist");
+                        last_child.branches.extend(branches);
+                        last_child.children.extend(grandchildren);
+
+                        graph.remove_child(current_id, *child_id);
+                    }
+
+                    enqueue_merge_stack(&mut queue, graph, last_child_id);
+                }
+            }
+        }
+    }
+}
+
+fn enqueue_merge_stack(
+    queue: &mut VecDeque<(git2::Oid, CommitTimesByTreeId)>,
+    graph: &Graph,
+    node_id: git2::Oid,
+) {
+    let mut unprotected_children = CommitTimesByTreeId::new();
+    for child_id in graph
+        .get(node_id)
+        .expect("all children exist")
+        .children
+        .iter()
+        .copied()
+    {
+        let child = graph.get(child_id).expect("all children exist");
+
+        unprotected_children
+            .entry(child.commit.tree_id)
+            .or_insert_with(Vec::new)
+            .push((child.commit.time, child_id));
+    }
+    if !unprotected_children.is_empty() {
+        queue.push_back((node_id, unprotected_children));
+    }
+}
+
 pub fn to_script(graph: &Graph) -> crate::git::Script {
     let mut script = crate::git::Script::new();
 
@@ -691,8 +891,21 @@ pub fn to_script(graph: &Graph) -> crate::git::Script {
         let current = graph.get(current_id).expect("all children exist");
 
         for child_id in current.children.iter().copied() {
-            let child_action = graph.get(child_id).expect("all children exist").action;
+            let child = graph.get(child_id).expect("all children exist");
+            let child_action = child.action;
             if child_action.is_protected() {
+                if !child.branches.is_empty() {
+                    // We might be updating protected branches as part of a `pull --rebase`,
+                    let stack_mark = child.commit.id;
+                    script
+                        .commands
+                        .push(crate::git::Command::SwitchCommit(stack_mark));
+                    for branch in child.branches.iter() {
+                        script
+                            .commands
+                            .push(crate::git::Command::CreateBranch(branch.name.clone()));
+                    }
+                }
                 protected_queue.push_back(child_id);
             } else if let Some(mut dependent) = node_to_script(graph, child_id) {
                 dependent
@@ -733,12 +946,12 @@ fn node_to_script(graph: &Graph, node_id: git2::Oid) -> Option<crate::git::Scrip
                 extend_dependents(node, &mut script, node_dependents, transaction);
             }
         }
-        crate::graph::Action::Squash => {
+        crate::graph::Action::Fixup => {
             script
                 .commands
-                .push(crate::git::Command::Squash(node.commit.id));
+                .push(crate::git::Command::Fixup(node.commit.id));
             // We can't re-target the branches of the commit we are squashing into, so the ops that
-            // creates a `Squash` option has to handle that.
+            // creates a `Fixup` option has to handle that.
             for branch in node.branches.iter() {
                 script
                     .commands
@@ -764,11 +977,17 @@ fn node_to_script(graph: &Graph, node_id: git2::Oid) -> Option<crate::git::Scrip
                 .copied()
                 .filter_map(|child_id| node_to_script(graph, child_id))
                 .collect();
-            if !node_dependents.is_empty() {
+            if !node_dependents.is_empty() || !node.branches.is_empty() {
                 let stack_mark = node.commit.id;
                 script
                     .commands
                     .push(crate::git::Command::SwitchCommit(stack_mark));
+                // We might be updating protected branches as part of a `pull --rebase`,
+                for branch in node.branches.iter() {
+                    script
+                        .commands
+                        .push(crate::git::Command::CreateBranch(branch.name.clone()));
+                }
 
                 // No transactions needed for protected commits
                 let transaction = false;
